import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Matern

from active_bo_ros.PolicyModel.GaussianRBFModel import GaussianRBF
from active_bo_ros.AcquisitionFunctions.ExpectedImprovement import ExpectedImprovement
from active_bo_ros.AcquisitionFunctions.ProbabilityOfImprovement import ProbabilityOfImprovement
from active_bo_ros.AcquisitionFunctions.ConfidenceBound import ConfidenceBound


class BayesianOptimization:
    def __init__(self, env, nr_steps, nr_init=3, acq='ei', nr_weights=6, policy_seed=None):
        self.env = env
        self.nr_init = nr_init
        self.acq = acq

        self.X = None
        self.Y = None
        self.GP = None

        self.episode = 0
        self.counter_array = np.empty((1, 1))
        self.best_reward = np.empty((1, 1))

        self.nr_policy_weights = nr_weights
        self.nr_steps = nr_steps
        self.policy_seed = policy_seed
        self.lower_bound = -1.0
        self.upper_bound = 1.0

        self.policy_model = GaussianRBF(self.nr_policy_weights,
                                        self.nr_steps,
                                        self.policy_seed,
                                        self.lower_bound,
                                        self.upper_bound)

        self.eval_X = 100

    def reset_bo(self):
        self.counter_array = np.empty((1, 1))
        self.GP = None
        self.episode = 0
        self.best_reward = np.empty((1, 1))

    def runner(self, policy, seed=None):
        env_reward = 0.0
        step_count = 0

        self.env.reset(seed=seed)

        for i in range(len(policy)):
            action = policy[i]
            action_clipped = action.clip(min=-1.0, max=1.0)
            output = self.env.step(action_clipped.astype(np.float32))

            env_reward += output[1]
            done = output[2]
            step_count += 1

            if done:
                self.counter_array = np.vstack((self.counter_array, step_count))
                break

            if not done and i == len(policy):
                self.counter_array = np.vstack((self.counter_array, step_count))

        self.env.reset(seed=seed)
        return env_reward, step_count

    def initialize(self, seed=None):
        self.env.reset(seed=seed)
        self.reset_bo()

        self.X = np.zeros((self.nr_init, self.nr_policy_weights))
        self.Y = np.zeros((self.nr_init, 1))

        for i in range(self.nr_init):
            self.policy_model.random_policy()
            self.X[i, :] = self.policy_model.weights.T
            policy = self.policy_model.rollout()

            reward, step_count = self.runner(policy)

            self.Y[i] = reward

        self.GP = GaussianProcessRegressor(Matern(nu=1.5, length_scale_bounds=(1e-8, 1e5)), n_restarts_optimizer=5, )
        self.GP.fit(self.X, self.Y)

    def next_observation(self):
        if self.acq == "Expected Improvement":
            x_next = ExpectedImprovement(self.GP,
                                         self.X,
                                         self.eval_X,
                                         self.nr_policy_weights,
                                         kappa=0,
                                         seed=self.policy_seed,
                                         lower=self.lower_bound,
                                         upper=self.upper_bound)

        elif self.acq == "Probability of Improvement":
            x_next = ProbabilityOfImprovement(self.GP,
                                              self.X,
                                              self.eval_X,
                                              self.nr_policy_weights,
                                              kappa=0,
                                              seed=self.policy_seed,
                                              lower=self.lower_bound,
                                              upper=self.upper_bound)

        elif self.acq == "Upper Confidence Bound":
            x_next = ConfidenceBound(self.GP,
                                     self.eval_X,
                                     self.nr_policy_weights,
                                     beta=2.576,
                                     seed=self.policy_seed,
                                     lower=self.lower_bound,
                                     upper=self.upper_bound)

        else:
            raise NotImplementedError

        return x_next

    def eval_new_observation(self, x_next, seed=None):
        self.policy_model.weights = x_next
        policy = self.policy_model.rollout()

        reward, step_count = self.runner(policy, seed=seed)

        self.X = np.vstack((self.X, x_next))
        self.Y = np.vstack((self.Y, reward))

        self.GP.fit(self.X, self.Y)

        if self.episode == 0:
            self.best_reward[0] = max(self.Y)
        else:
            self.best_reward = np.vstack((self.best_reward, max(self.Y)))

        self.episode += 1
        return step_count

    def add_new_observation(self, reward, x_new):
        self.X = np.vstack((self.X, x_new))
        self.Y = np.vstack((self.Y, reward))

        self.GP.fit(self.X, self.Y)

        if self.episode == 0:
            self.best_reward[0] = max(self.Y)
        else:
            self.best_reward = np.vstack((self.best_reward, max(self.Y)))

        self.episode += 1

    def get_best_result(self):
        y_max = np.max(self.Y)
        idx = np.argmax(self.Y)
        x_max = self.X[idx, :]

        self.policy_model.weights = x_max
        best_policy = self.policy_model.rollout().reshape(-1,)

        return best_policy, y_max, x_max
